# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import brax
from brax.io import html
from IPython.display import HTML

from deluca.core import Env
from deluca.core import field


class BraxEnv(Env):
    sys: brax.System = field(jaxed=False)

    def setup(self):
        if isinstance(self.sys, brax.physics.config_pb2.Config):
            self.sys = brax.System(self.sys)

    def init(self):
        return self.sys.default_qp()

    def __call__(self, state, action):
        state, _ = self.sys.step(state, action)
        return state, state

    def render(self, states):
        return HTML(html.render(self.sys, states))
